//
// file: svsglwin.cpp
//

#include <stdio.h>
#include <math.h>
#include "flwin.h"
#include "svsglwin.h"
#include <FL/Fl.H>
#include "glframe.h"
//
// GL frame interface
//

glframe *maingl = NULL; 	// 3D display window

void Display3D(svsStereoImage *di, svs3Dpoint *globalPts3D)
{
  	  static int w = 425, h = 340;
	  int xx, yy, ww, hh;
	  if (!maingl)
	  {
		maingl = new glframe();
		svsSleep(500);		// GDI32 seems to need some time to set up
	  }

	  xx = maingl->gl_win->x();
	  yy = maingl->gl_win->y();
	  ww = maingl->gl_win->w();
	  hh = maingl->gl_win->h();

	  if (w != ww || h != hh)	// resize, have to destroy and re-create to get it to show...
		{
		//      debugMessage("Setting up GL window");
		  maingl->win_3d->begin();
		  delete maingl->gl_win;
		  SvsGLWindow* o = maingl->gl_win = new SvsGLWindow(xx, yy, ww, hh);
		  w = ww; h = hh;
		  o->box(FL_EMBOSSED_FRAME);
		  Fl_Group::current()->resizable(o);
		  maingl->win_3d->end();
		}

	  maingl->gl_win->DisplayDisparityImage(di, globalPts3D);
	  maingl->win_3d->show();

	  maingl->gl_win->show();

	  Fl::run();
}


//
// GL graphics window definitions
//

SvsGLWindow::SvsGLWindow(int X, int Y, int W, int H) 
  : Fl_Gl_Window(X, Y, W, H) 
{

  // effects: SvsGLWindow contstructor; creates the window and
  //          initializes the class variables

  anglex = 0;      // in degrees
  angley = 180;    // in degrees
  anglez = 0;      // in degrees
  newModel = 0;    // should we recalculate a new model view?
  moving = 0;      // only needed for mouse interaction (disabled for now)
  numPoints = bufsize = 0;   // number of 3D points we get from a disparity image
  axes = 1;        // display or don't display the axes

  scale = 1.0;     // don't scale for now
  scaleFactor = 4.0;
  xshift = 0.0;    // no shifts for now
  yshift = 0.0;
  zshift = 0.0;

  centerx = centery = 0;
  centerz = 1.0;
}
  
void SvsGLWindow::setDisplayAxes(int value)
{
  // requires: value = 0 or 1
  //  effects: sets the axes value to value.
  //           if value = 0, does not display axes on image.
  //           if value = 1, displays axes on image.

  axes = value;
  redraw();
}

void SvsGLWindow::scaleImage(float value)
{
  // requires: 0 <= value <= 1
  //  effects: scales the image according to an arbitrary
  //           constant (4) times value.

  scale = value * scaleFactor;
  newModel = 1;
  redraw();
}

void SvsGLWindow::transImageH(float value)
{
  // requires: 0 <= value <= 1
  //  effects: translates the image along the "horizontal"
  //           axis; a value of 0.5 is zero translation.

  xshift = (float)(0.5-value) * 8;
  newModel = 1;
  redraw();
}

void SvsGLWindow::transImageV(float value)
{
  // requires: 0 <= value <= 1
  //  effects: translates the image along the "vertical"
  //           axis; a value of 0.5 is zero translation.

  yshift = (float)(0.5-value) * 8;
  newModel = 1;
  redraw();
}

void SvsGLWindow::transImageZ(float value)
{
  // requires: 0 <= value <= 1
  //  effects: translates the image along the "z (depth)"
  //           axis; a value of 0.5 is zero translation.

  zshift = (float)(0.5-value) * 20;
  newModel = 1;
  redraw();
}

void SvsGLWindow::rotImageH(float value)
{
  // requires: 0 <= value <= 1
  //  effects: rotates the image around the "horizontal"
  //           axis; a value of 0.5 is zero rotation.

  anglex = (float)(0.5-value)*540;
  newModel = 1;
  redraw();

}

void SvsGLWindow::rotImageV(float value)
{
  // requires: 0 <= value <= 1
  //  effects: rotates the image around the "vertical"
  //           axis; a value of 0.5 is zero rotation.

  angley = (float)(0.5-value)*540 + 180;
  newModel = 1;
  redraw();
}

void SvsGLWindow::rotImageZ(float value)
{
  // requires: 0 <= value <= 1
  //  effects: rotates the image around the "z (depth)"
  //           axis; a value of 0.5 is zero rotation.

  anglez = (float)(0.5 - value)*540;
  newModel = 1;
  redraw();
}



// calcCenterZ
// effects: takes a disparity image and calculates average depth
// in a pixel neighborhood nh.
// assumes 3D has already been calculated

static GLfloat gamma_table[256];    // gamma correction array

bool SvsGLWindow::calcCenterZ(svsStereoImage *di, int nh, svs3Dpoint* globalPts3D)
{
  int w, h;
  
  
  //svs3Dpoint *pts = di->pts3D;
  svs3Dpoint *pts = globalPts3D;
  
  svs3Dpoint *pt;
  int np = 0;
  int ld, td;

  w = di->ip.width;   
  h = di->ip.height;
 
  // set center neighborhood boundaries
  ld = w/2 - nh/2;
  if (ld < 0)
    ld = 0;
  td = h/2 - nh/2;
  if (td < 0)
    td = 0;

  if (ld + nh > w)
    nh = w - ld - 1;
  if (td + nh > h)
    nh = h - td - 1;

  centerz = 0.0;
  float minz = 100000.0;
  int i, j;
  
  for (j=td; j<td+nh; j++)
    {
      pt = pts + j*w + ld;
      for (i=ld; i<ld+nh; i++, pt++)
	{
	  if (pt->A > 0.0) 
	    {
	      centerz = centerz + pt->Z;
              if (pt->Z < minz)
                minz = pt->Z;
	      np++;
	    }
	}
    }

  if (np > 0)
    {
      centerz = centerz/np;
      centerz = minz;		// try the minimum...
      if (centerz > 10)
	centerz = 10;
      return true;
    }
  return false;
}


// 
// Uses the 3D reconstruction and monochrome or color image
// to display a 3D rendering of the stereo image
//

void SvsGLWindow::DisplayDisparityImage(svsStereoImage *di, svs3Dpoint* globalPts3D)
{
  int w, h;
  int ip = 0, cp;
  svs3Dpoint *pts;
  unsigned char *c = NULL;
  svsColorPixel *cc = NULL; 
  unsigned char colorR, colorG, colorB;
  static double old_gamma = -1;
  double gam;
  bool cflag;

  //  debugMessage("Display disparity image");

  // calculate gamma table if needed
  gam = di->ip.gamma;
  if (gam <= 0.0)
    gam = 1.0;
  if (gam >= 1.0)
    gam = 1.0;
  if (gam != old_gamma)		// set up gamma array
    {
      old_gamma = gam;
      for (int i=0; i<256; i++)
	gamma_table[i] = (GLfloat)(pow((double)i/255.0, gam));
    }

  w = di->ip.width;
  h = di->ip.height;

  numPoints = di->numPoints;
  if (numPoints > bufsize)	// check buffers
    {
      if (bufsize > 0)
	{
	  delete [] pointListX;
	  delete [] pointListY;
	  delete [] pointListZ;
	  delete [] colorListR;
	  delete [] colorListG;
	  delete [] colorListB;
	}
      bufsize = numPoints;
      pointListX = new GLfloat[bufsize];
      pointListY = new GLfloat[bufsize];
      pointListZ = new GLfloat[bufsize];
      colorListR = new GLfloat[bufsize];
      colorListG = new GLfloat[bufsize];
      colorListB = new GLfloat[bufsize];
    }
 
  cflag = di->haveColor;
  if (cflag)
    cc = (svsColorPixel *)di->color;
  else
    c = di->left;

  maxx = maxy = maxz = -1000;
  minx = miny = minz = 1000;

  int i, j;
  for (j=0; j<h; j++)
    {
      cp =  j*w;
      
      //pts = di->pts3D + cp;	// 3D point array
      
      pts = globalPts3D + cp;	// 3D point array
      
      for (i=0; i<w; i++, cp++, pts++)
	{
	  if (pts->A > 0)	// invalid points have A = 0, -1, -2
	    {
	      pointListX[ip] = pts->X;
	      pointListY[ip] = pts->Y;
	      pointListZ[ip] = pts->Z;

/*
		if(pts->Y < 0)
		{
			colorR = 255;
		  	colorB = 0;
		  	colorG = 0;
		  	colorListR[ip] = gamma_table[colorR];
		  	colorListG[ip] = gamma_table[colorG];
		 	colorListB[ip] = gamma_table[colorB];
		}
		else
		{

*/
		//------------------------------------

	      if (!cflag)	// monochrome
		{
		  colorR = c[cp];
		  colorListR[ip] = colorListG[ip] = colorListB[ip] = gamma_table[colorR];
		}
	      else 
		{
		  colorR = cc[cp].r;
		  colorB = cc[cp].b;
		  colorG = cc[cp].g;
		  colorListR[ip] = gamma_table[colorR];
		  colorListG[ip] = gamma_table[colorG];
		  colorListB[ip] = gamma_table[colorB];
		}
		

		//------------------------------------
/*		}
*/




              ip++;
	    }
	}
    }

  numPoints = ip;

  for (i=20; i<100; i+=10)
    {
      if (calcCenterZ(di,i,globalPts3D))
	break;
      else
	centerz = 1.0;
    }

  //  debugMessage("displaying %d points", numPoints);
  //  debugMessage("center: (%f, %f, %f)", centerx, centery, centerz);
  //  debugMessage("%d/%d points, center (0,0,%d)", numPoints, di->numPoints, (int)centerz);
//  debugMessage("extremes: ([%f, %f], [%f, %f], [%f, %f])", maxx, minx,
//	 maxy, miny, maxz, minz);

  newModel = 1;
  redraw();
}

void SvsGLWindow::recalcModelView() 
{
  // effects: recalculates the view of the model based on
  //          class variables (angle, scale, shift, center).
  //
  // rotates just around the centerpoint
  // translates from viewpoint
 
  glPopMatrix();
  glPushMatrix();

  glTranslatef(xshift, yshift, zshift);

  glRotatef(anglex, 0.0, 1.0, 0.0);
  glRotatef(angley, 1.0, 0.0, 0.0);
  glRotatef(anglez, 0.0, 0.0, 1.0);
  
  glScalef(scale, scale, scale);

//  glTranslatef(xshift*1000, yshift*1000, zshift*1000);

  glTranslatef(-centerx, -centery, -centerz);

  newModel = 0;  
}

void SvsGLWindow::draw() 
{
  // effects: draw function called when necessary to update
  //          the OpenGL window.
  //          draws the 3D points and axes.
  
  int i;

  if (!valid()) {
    // ... set up projection, viewport, etc ...
    // ... window size is in w() and h().
    // ... valid() is turned on by FLTK after draw() returns

    // hmm, should this be here?? or in an init somewhere?
	glLoadIdentity();
    glShadeModel(GL_FLAT);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_POINT_SMOOTH);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    glMatrixMode(GL_PROJECTION);
    gluPerspective(40.0,  // field of view in degree
		   1.0,   // aspect ratio  
		   0.5,   // Z near 
		   100.0);// Z far
    glMatrixMode(GL_MODELVIEW);
#if 0
    gluLookAt(0.0, 1.0, 8.0,  // eye location
	      0.0, 1.0, 0.0,  // center is at (0,0,0)
	      0.0, 1.0, 0.);  // up is in postivie Y direction 
#endif
    gluLookAt(0.5, 0.5, 4.0,  // eye location
	      0.5, 0.5, 0.0,  // what we look at
	      0.0, 1.0, 0.0); // up is in postive Y direction 
    glPushMatrix();    // dummy push so we can pop on model recalc
    
    glPointSize(1);
  }

  // ... draw ...

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  if(newModel)
    recalcModelView();

  glDisable(GL_TEXTURE_2D);


  glBegin(GL_POINTS);
  for(i=0; i<numPoints; i++) {
    glColor3f(colorListR[i], colorListG[i], colorListB[i]);
    glVertex3f(pointListX[i], pointListY[i], pointListZ[i]);
  }
  glEnd();

  if(axes) 
    {
      glBegin(GL_LINES);
      glColor3f(0, 0, 1);
      glVertex3f(0, 0, 0);
      glVertex3f(1.0, 0, 0);
      glEnd();
    
      glBegin(GL_LINES);
      glColor3f(0, 1, 0);
      glVertex3f(0, 0, 0);
      glVertex3f(0, 1.0, 0);
      glEnd();

      // this is the camera optical ray
      glBegin(GL_LINES);
      glColor3f(1, 0, 0);
      glVertex3f(0, 0, 0);
      glVertex3f(0, 0, 20.0);
      glEnd();

#if 0
      // show centerpoint
      glBegin(GL_LINES);
      glColor3f(0, 0, 1);
      glVertex3f(centerx, centery, centerz);    
      glVertex3f(centerx, centery, centerz+1000);
      glEnd();
#endif
    }



}


int SvsGLWindow::handle(int event) 
{
  // effects: handles users events in the gl window
  // rotates the 3D model

  switch(event) {
    
  case FL_PUSH:
    // mouse down event
    moving = 1;
    beginx = Fl::event_x();
    beginy = Fl::event_y();
    return 1;
  case FL_DRAG:
    // mouse drag (while down) event
    if(moving) {
      anglex = anglex + (Fl::event_x()-beginx);
      beginx = Fl::event_x();
      angley = angley + (Fl::event_y()-beginy);
      beginy = Fl::event_y();
      newModel = 1;
      redraw();
    }
    return 1;
  case FL_RELEASE:
    // mouse up event
    moving = 0;
    return 1;
  default: 
    return Fl_Gl_Window::handle(event);
    
  }
}



//
// callbacks for OpenGL window
//

void
gl_reset_cb(Fl_Button *w, void *x)
{
  // callback: on "Reset" button in 3D Display window
  //  effects: resets the view to the default

  maingl->zoom_val->value(0.25);
  maingl->gl_win->scaleImage(0.25);
  maingl->vert_trans_val->value(0.5);
  maingl->gl_win->transImageV(0.5);
  maingl->horz_trans_val->value(0.5);
  maingl->gl_win->transImageH(0.5);
  maingl->gl_win->rotImageV(0.5);
  maingl->gl_win->rotImageH(0.5);
  maingl->z_ang_val->value(0.5);
  maingl->gl_win->rotImageZ(0.5);
  maingl->gl_win->transImageZ(0.5);


}

// 3D Display rotation, translation, and zoom sliders/rollers

void
gl_axes_cb(Fl_Light_Button *w, void *x)
{
  maingl->gl_win->setDisplayAxes(w->value());
}

//
// this doesn't work...
//

void
gl_texture_cb(Fl_Light_Button *w, void *x)
{
}

void
gl_zoom_cb(Fl_Slider *w, void *x)
{
  maingl->gl_win->scaleImage((float)w->value());
}

void 
gl_vert_trans_cb(Fl_Slider *w, void *x)
{
  maingl->gl_win->transImageV((float)w->value());
}

void
gl_vert_ang_cb(Fl_Roller *w, void *x)
{
  maingl->gl_win->rotImageV((float)w->value());
}

void
gl_z_ang_cb(Fl_Roller *w, void *x)
{
  maingl->gl_win->rotImageZ((float)w->value());
}

void
gl_horz_trans_cb(Fl_Slider *w, void *x)
{
  maingl->gl_win->transImageH((float)w->value());
}

void
gl_horz_ang_cb(Fl_Roller *w, void *x)
{
  maingl->gl_win->rotImageH((float)w->value());
}

void
gl_z_trans_cb(Fl_Slider *w, void *x)
{
  maingl->gl_win->transImageZ((float)w->value());
}
